# Unix SMB/CIFS implementation.
#
# Copyright (C) Catalyst.Net Ltd. 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Test whether various python calls segfault when given unexpected input.
"""

import samba.tests
import os
import sys
from samba.net import Net
from samba.credentials import DONT_USE_KERBEROS
from samba.dcerpc import misc, drsuapi, samr, unixinfo, dnsserver
from samba import auth, gensec
from samba.samdb import SamDB
from samba import netbios
from samba import registry
from samba import ldb
from samba import messaging
from samba import dsdb

import traceback


def segfault_detector(f):
    def wrapper(*args, **kwargs):
        pid = os.fork()
        if pid == 0:
            try:
                f(*args, **kwargs)
            except Exception as e:
                traceback.print_exc()
            sys.stderr.flush()
            sys.stdout.flush()
            os._exit(0)

        pid2, status = os.waitpid(pid, 0)
        if os.WIFSIGNALED(status):
            signal = os.WTERMSIG(status)
            raise AssertionError("Failed with signal %d" % signal)

    return wrapper


def no_gdb_backtrace(f):
    from os import environ
    def w(*args, **kwargs):
        old = environ.get('PLEASE_NO_GDB_BACKTRACE')
        environ['PLEASE_NO_GDB_BACKTRACE'] = '1'
        try:
            f(*args, **kwargs)
        finally:
            if old is not None:
                environ['PLEASE_NO_GDB_BACKTRACE'] = old
            else:
                del environ['PLEASE_NO_GDB_BACKTRACE']

    return w


class SegfaultTests(samba.tests.TestCase):
    def get_lp_et_al(self):
        server = os.environ["SERVER"]
        lp = self.get_loadparm()

        creds = self.insta_creds(template=self.get_credentials(),
                                 kerberos_state=DONT_USE_KERBEROS)
        return lp, creds, server

    def get_samdb(self):
        lp, creds, server = self.get_lp_et_al()
        url = 'ldap://' + server
        ldb = SamDB(url, credentials=creds, lp=lp)
        return ldb

    @segfault_detector
    def test_net_replicate_init__1(self):
        lp, creds, server = self.get_lp_et_al()
        net = Net(creds, lp, server=server)
        net.replicate_init(42, lp, None, misc.GUID())

    @no_gdb_backtrace
    @segfault_detector
    def test_net_replicate_init__3(self):
        # third argument is also unchecked
        samdb = self.get_samdb()
        lp, creds, server = self.get_lp_et_al()
        net = Net(creds, lp, server=server)
        net.replicate_init(samdb, lp, 42, misc.GUID())

    @segfault_detector
    def test_net_replicate_chunk_1(self):
        lp, creds, server = self.get_lp_et_al()
        ctr = drsuapi.DsGetNCChangesCtr6()
        net = Net(creds, lp, server=server)
        net.replicate_chunk(42, 1, ctr)

    @segfault_detector
    def test_auth_context_gensec_start_server(self):
        a = auth.AuthContext(ldb=42, methods=['sam'])
        # there is no failure yet because the ldb is not actually
        # dereferenced.
        g = gensec.Security.start_server(auth_context=a)
        # and still the ldb is not dereferenced...

    @segfault_detector
    def test_auth_user_session(self):
        s = auth.user_session(ldb=42, principal='foo')

    @segfault_detector
    def test_gensec_start_server(self):
        gensec.Security.start_server(auth_context=42)

    @segfault_detector
    def test_netbios_query_name(self):
        n = netbios.Node()
        t = n.query_name((42, 'foo'), 'localhost')

    @segfault_detector
    def test_encrypt_netr_crypt_password(self):
        lp, creds, server = self.get_lp_et_al()
        creds.encrypt_netr_crypt_password(42)

    @segfault_detector
    def test_hive_open_ldb(self):
        # we don't need to provide a valid path because we segfault first
        try:
            registry.open_ldb('', credentials=42)
        except ldb.LdbError as e:
            print("failed with %s" % e)

    @segfault_detector
    def test_hive_open_hive(self):
        # we don't need to provide a valid path because we segfault first
        try:
            registry.open_hive('s', 's', 's', 's')
        except ldb.LdbError as e:
            print("failed with %s" % e)

    @segfault_detector
    def test_ldb_add_nameless_element(self):
        m = ldb.Message()
        e = ldb.MessageElement('q')
        try:
            m.add(e)
        except ldb.LdbError:
            pass
        str(m)

    @segfault_detector
    def test_messaging_deregister(self):
        messaging.deregister('s', 's', 's', False)

    @segfault_detector
    def test_rpcecho(self):
        from samba.dcerpc import echo
        echo.rpcecho("")

    @segfault_detector
    def test_dcerpc_idl_ref_elements(self):
        """There are many pidl generated functions that crashed on this
        pattern, where a NULL pointer was created rather than an empty
        structure."""
        samr.Connect5().out_info_out = 1

    @segfault_detector
    def test_dcerpc_idl_unixinfo_elements(self):
        """Dereferencing is sufficient to crash"""
        unixinfo.GetPWUid().out_infos

    @segfault_detector
    def test_dcerpc_idl_inline_arrays(self):
        """Inline arrays were incorrectly handled."""
        dnsserver.DNS_RPC_SERVER_INFO_DOTNET().pExtensions

    @segfault_detector
    def test_dcerpc_idl_set_inline_arrays(self):
        """Setting an inline array was incorrectly handled."""
        a = dnsserver.DNS_EXTENSION()
        x = dnsserver.DNS_RPC_DP_INFO()
        x.pwszReserved = [a, a, a]

    @no_gdb_backtrace
    @segfault_detector
    def test_dnsp_string_list(self):
        from samba.dcerpc import dnsp
        # We segfault if s.count is greater than the length of s.str
        s = dnsp.string_list()
        s.count = 3
        s.str

    @no_gdb_backtrace
    @segfault_detector
    def test_dns_record(self):
        from samba.dnsserver import TXTRecord
        from samba.dcerpc import dnsp
        # there are many others here
        rec = TXTRecord(["a", "b", "c"])
        rec.wType = dnsp.DNS_TYPE_A
        rec.data

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_msg_diff(self):
        samdb = self.get_samdb()

        msg = ldb.Message()
        msg.dn = ldb.Dn(samdb, '')
        diff = samdb.msg_diff(msg, msg)

        del msg
        diff.dn

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_msg_del_dn(self):
        msg = ldb.Message()
        del msg.dn

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_control_del_critical(self):
        samdb = self.get_samdb()

        c = ldb.Control(samdb, 'relax:1')
        del c.critical

    @segfault_detector
    def test_random_bytes(self):
        # memory error from SIZE_MAX -1 allocation.
        from samba import generate_random_bytes
        generate_random_bytes(-1)

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_get_default_basedn(self):
        samdb = self.get_samdb()

        dn = samdb.get_default_basedn()
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        del samdb
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_get_ncroot_existing(self):
        samdb = self.get_samdb()

        base_dn = samdb.get_default_basedn()
        dn = samdb.get_nc_root(base_dn)
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        del samdb
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_get_ncroot_not_existing(self):
        samdb = self.get_samdb()

        base_dn = samdb.get_default_basedn()
        base_dn.add_child("CN=TEST")
        dn = samdb.get_nc_root(base_dn)
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        del samdb
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_get_wellknown_dn(self):
        samdb = self.get_samdb()

        base_dn = samdb.get_default_basedn()
        wk_dn = samdb.get_wellknown_dn(base_dn, dsdb.DS_GUID_LOSTANDFOUND_CONTAINER)
        wk_dn.add_child("CN=TEST")
        wk_dn.set_component(0, "CN", "Test2")
        del samdb
        del base_dn
        wk_dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_concat(self):
        samdb = self.get_samdb()

        dn1 = ldb.Dn(samdb, "CN=foo")
        dn2 = ldb.Dn(samdb, "CN=bar")
        dn = dn1 + dn2
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        del samdb
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_get_parent(self):
        samdb = self.get_samdb()

        dn1 = ldb.Dn(samdb, "CN=foo,CN=bar")
        dn = dn1.parent()
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        del samdb
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_search_result(self):
        samdb = self.get_samdb()
        base_dn = samdb.get_default_basedn()
        res = samdb.search(base=base_dn,
                           scope=ldb.SCOPE_SUBTREE,
                           attrs=[],
                           expression="(cn=administrator)")
        msg = res[0]
        dn1 = msg.dn
        dn = dn1.parent()
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        del samdb
        del msg
        del dn1
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")

        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")

        del samdb
        del msg
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_ldif_parse(self):

        ldif = """dn: cn=test
        changetype: add
        -
        cn: test
        -
        """
        samdb = self.get_samdb()
        for changetype, msg in samdb.parse_ldif(ldif):
            dn = msg.dn
            dn.add_child("CN=TEST")
            dn.set_component(0, "CN", "Test2")

            del samdb
            del msg
            dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_dict_init(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")

        msg2 = ldb.Message.from_dict(samdb,
                                     {"dn": msg.dn,
                                      "foo": ["bar"]})
        del msg
        dn = msg2.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")

        del samdb

        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_msg_init(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")

        msg2 = ldb.Message(dn=msg.dn)
        del msg
        dn = msg2.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")

        del samdb

        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_msg_diff(self):
        samdb = self.get_samdb()
        msg = ldb.Message()
        msg.dn = ldb.Dn(samdb, "CN=foo")
        msg["foo"] = ["bar"]

        msg2 = ldb.Message()
        msg2.dn = ldb.Dn(samdb, "CN=foo")
        msg2["foo"] = ["bar2"]

        msg3 = samdb.msg_diff(msg, msg2)
        dn = msg3.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")

        del samdb
        del msg3
        del msg2
        del msg
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_msg_from_dict(self):
        samdb = self.get_samdb()
        msg = ldb.Message.from_dict(samdb,
                                    {
                                        "dn": "CN=foo",
                                        "foo": ["bar"]},
                                    ldb.FLAG_MOD_REPLACE)

        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")

        del samdb
        del msg
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign_disconnecting_connection(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")
        samdb.disconnect()
        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")

        del samdb
        del msg
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign_disconnecting_connection_no_del(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")
        samdb.disconnect()
        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign_disconnecting_connection_later__no_del(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")
        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        samdb.disconnect()
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign_disconnecting_connection_add_child_later(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")
        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        samdb.disconnect()
        dn.add_child("CN=TeSt")

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign_disconnecting_connection_later(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")
        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        samdb.disconnect()
        del samdb
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign_disconnecting_connection_reconnecting_later(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")
        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        samdb.disconnect()
        lp, creds, server = self.get_lp_et_al()
        url = 'ldap://' + server
        samdb.set_loadparm(lp)
        samdb.connect(url)
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign_switching_out_connection(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")
        lp, creds, server = self.get_lp_et_al()
        url = 'ldap://' + server
        samdb.disconnect()
        samdb.set_loadparm(lp)
        samdb.connect(url)

        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")
        samdb.disconnect()
        samdb.set_loadparm(lp)
        samdb.connect(url)
        del samdb
        del msg
        dn.get_casefold()

    @no_gdb_backtrace
    @segfault_detector
    def test_ldb_use_after_free_dn_assign_disconnecting_and_switching_out_connection(self):

        msg = ldb.Message()

        samdb = self.get_samdb()
        msg.dn = ldb.Dn(samdb, "CN=Test")
        samdb.disconnect()
        lp, creds, server = self.get_lp_et_al()
        url = 'ldap://' + server
        samdb.set_loadparm(lp)
        samdb.connect(url)

        dn = msg.dn
        dn.add_child("CN=TEST")
        dn.set_component(0, "CN", "Test2")

        del samdb
        del msg
        dn.get_casefold()



#!/usr/bin/env python3
# Simple tests for the ldb python bindings.
# Copyright (C) 2007 Jelmer Vernooij <jelmer@samba.org>

import os
import sys
sys.path.insert(0, "bin/python")
from api_base import (
    MDB_PREFIX,
    MDB_INDEX_OBJ,
    tempdir,
    LdbBaseTest
)

import ldb
import shutil


class SimpleLdb(LdbBaseTest):

    def setUp(self):
        super().setUp()
        self.testdir = tempdir()
        self.filename = os.path.join(self.testdir, "test.ldb")
        self.ldb = ldb.Ldb(self.url(), flags=self.flags())
        try:
            self.ldb.add(self.index)
        except AttributeError:
            pass

    def tearDown(self):
        self.ldb.disconnect()
        shutil.rmtree(self.testdir)
        super().tearDown()
        # Ensure the LDB is closed now, so we close the FD

    def test_connect(self):
        ldb.Ldb(self.url(), flags=self.flags())

    def test_connect_none(self):
        ldb.Ldb()

    def test_connect_later(self):
        x = ldb.Ldb()
        x.connect(self.url(), flags=self.flags())

    def test_connect_twice(self):
        url = self.url()
        x = ldb.Ldb(url)
        with self.assertRaises(ldb.LdbError):
            x.connect(url, flags=self.flags())

    def test_connect_twice_later(self):
        url = self.url()
        flags = self.flags()
        x = ldb.Ldb()
        x.connect(url, flags)
        with self.assertRaises(ldb.LdbError):
            x.connect(url, flags)

    def test_connect_and_disconnect(self):
        url = self.url()
        flags = self.flags()
        x = ldb.Ldb()
        x.connect(url, flags)
        x.disconnect()
        x.connect(url, flags)
        x.disconnect()

    def test_repr(self):
        x = ldb.Ldb()
        self.assertTrue(repr(x).startswith("<ldb connection"))

    def test_set_create_perms(self):
        x = ldb.Ldb()
        x.set_create_perms(0o600)

    def test_search(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(len(l.search()), 0)

    def test_search_controls(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(len(l.search(controls=["paged_results:0:5"])), 0)

    def test_utf8_ldb_Dn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        dn = ldb.Dn(l, (b'a=' + b'\xc4\x85\xc4\x87\xc4\x99\xc5\x82\xc5\x84\xc3\xb3\xc5\x9b\xc5\xba\xc5\xbc').decode('utf8'))

    def test_utf8_encoded_ldb_Dn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        dn_encoded_utf8 = b'a=' + b'\xc4\x85\xc4\x87\xc4\x99\xc5\x82\xc5\x84\xc3\xb3\xc5\x9b\xc5\xba\xc5\xbc'
        try:
            dn = ldb.Dn(l, dn_encoded_utf8)
        except UnicodeDecodeError as e:
                raise
        except TypeError as te:
            p3errors = ["argument 2 must be str, not bytes",
                        "Can't convert 'bytes' object to str implicitly"]
            self.assertIn(str(te), p3errors)

    def test_search_attrs(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(len(l.search(ldb.Dn(l, ""), ldb.SCOPE_SUBTREE, "(dc=*)", ["dc"])), 0)

    def test_search_string_dn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(len(l.search("", ldb.SCOPE_SUBTREE, "(dc=*)", ["dc"])), 0)

    def test_search_attr_string(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertRaises(TypeError, l.search, attrs="dc")
        self.assertRaises(TypeError, l.search, attrs=b"dc")

    def test_opaque(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        l.set_opaque("my_opaque", True)
        self.assertTrue(l.get_opaque("my_opaque") is not None)
        self.assertEqual(None, l.get_opaque("unknown"))

    def test_opaque_bool(self):
        """Test that we can set boolean opaque values."""

        db = ldb.Ldb(self.url(), flags=self.flags())
        name = "my_opaque"

        db.set_opaque(name, False)
        self.assertEqual(False, db.get_opaque(name))

        db.set_opaque(name, True)
        self.assertEqual(True, db.get_opaque(name))

    def test_opaque_int(self):
        """Test that we can set (positive) integer opaque values."""

        db = ldb.Ldb(self.url(), flags=self.flags())
        name = "my_opaque"

        db.set_opaque(name, 0)
        self.assertEqual(0, db.get_opaque(name))

        db.set_opaque(name, 12345678)
        self.assertEqual(12345678, db.get_opaque(name))

        # Negative values can’t be set.
        self.assertRaises(OverflowError, db.set_opaque, name, -99999)

    def test_opaque_string(self):
        """Test that we can set string opaque values."""

        db = ldb.Ldb(self.url(), flags=self.flags())
        name = "my_opaque"

        db.set_opaque(name, "")
        self.assertEqual("", db.get_opaque(name))

        db.set_opaque(name, "foo bar")
        self.assertEqual("foo bar", db.get_opaque(name))

    def test_opaque_none(self):
        """Test that we can set an opaque to None to effectively unset it."""

        db = ldb.Ldb(self.url(), flags=self.flags())
        name = "my_opaque"

        # An opaque that has not been set is the same as None.
        self.assertIsNone(db.get_opaque(name))

        # Give the opaque a value.
        db.set_opaque(name, 3)
        self.assertEqual(3, db.get_opaque(name))

        # Test that we can set the opaque to None to unset it.
        db.set_opaque(name, None)
        self.assertIsNone(db.get_opaque(name))

    def test_opaque_unsupported(self):
        """Test that trying to set unsupported values raises an error."""

        db = ldb.Ldb(self.url(), flags=self.flags())
        name = "my_opaque"

        self.assertRaises(ValueError, db.set_opaque, name, [])
        self.assertRaises(ValueError, db.set_opaque, name, ())
        self.assertRaises(ValueError, db.set_opaque, name, 3.14)
        self.assertRaises(ValueError, db.set_opaque, name, 3+2j)
        self.assertRaises(ValueError, db.set_opaque, name, b'foo')

    def test_search_scope_base_empty_db(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(len(l.search(ldb.Dn(l, "dc=foo1"),
                                      ldb.SCOPE_BASE)), 0)

    def test_search_scope_onelevel_empty_db(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(len(l.search(ldb.Dn(l, "dc=foo1"),
                                      ldb.SCOPE_ONELEVEL)), 0)

    def test_delete(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertRaises(ldb.LdbError, lambda: l.delete(ldb.Dn(l, "dc=foo2")))

    def test_delete_w_unhandled_ctrl(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=foo1")
        m["b"] = [b"a"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        self.assertRaises(ldb.LdbError, lambda: l.delete(m.dn, ["search_options:1:2"]))
        l.delete(m.dn)

    def test_contains(self):
        name = self.url()
        l = ldb.Ldb(name, flags=self.flags())
        self.assertFalse(ldb.Dn(l, "dc=foo3") in l)
        l = ldb.Ldb(name, flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=foo3")
        m["b"] = ["a"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        try:
            self.assertTrue(ldb.Dn(l, "dc=foo3") in l)
            self.assertFalse(ldb.Dn(l, "dc=foo4") in l)
        finally:
            l.delete(m.dn)

    def test_get_config_basedn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(None, l.get_config_basedn())

    def test_get_root_basedn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(None, l.get_root_basedn())

    def test_get_schema_basedn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(None, l.get_schema_basedn())

    def test_get_default_basedn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(None, l.get_default_basedn())

    def test_add(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=foo4")
        m["bla"] = b"bla"
        m["objectUUID"] = b"0123456789abcdef"
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        try:
            self.assertEqual(len(l.search()), 1)
        finally:
            l.delete(ldb.Dn(l, "dc=foo4"))

    def test_search_iterator(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        s = l.search_iterator()
        s.abandon()
        try:
            for me in s:
                self.fail()
            self.fail()
        except RuntimeError as re:
            pass
        try:
            s.abandon()
            self.fail()
        except RuntimeError as re:
            pass
        try:
            s.result()
            self.fail()
        except RuntimeError as re:
            pass

        s = l.search_iterator()
        count = 0
        for me in s:
            self.assertTrue(isinstance(me, ldb.Message))
            count += 1
        r = s.result()
        self.assertEqual(len(r), 0)
        self.assertEqual(count, 0)

        m1 = ldb.Message()
        m1.dn = ldb.Dn(l, "dc=foo4")
        m1["bla"] = b"bla"
        m1["objectUUID"] = b"0123456789abcdef"
        l.add(m1)
        try:
            s = l.search_iterator()
            msgs = []
            for me in s:
                self.assertTrue(isinstance(me, ldb.Message))
                count += 1
                msgs.append(me)
            r = s.result()
            self.assertEqual(len(r), 0)
            self.assertEqual(len(msgs), 1)
            self.assertEqual(msgs[0].dn, m1.dn)

            m2 = ldb.Message()
            m2.dn = ldb.Dn(l, "dc=foo5")
            m2["bla"] = b"bla"
            m2["objectUUID"] = b"0123456789abcdee"
            l.add(m2)

            s = l.search_iterator()
            msgs = []
            for me in s:
                self.assertTrue(isinstance(me, ldb.Message))
                count += 1
                msgs.append(me)
            r = s.result()
            self.assertEqual(len(r), 0)
            self.assertEqual(len(msgs), 2)
            if msgs[0].dn == m1.dn:
                self.assertEqual(msgs[0].dn, m1.dn)
                self.assertEqual(msgs[1].dn, m2.dn)
            else:
                self.assertEqual(msgs[0].dn, m2.dn)
                self.assertEqual(msgs[1].dn, m1.dn)

            s = l.search_iterator()
            msgs = []
            for me in s:
                self.assertTrue(isinstance(me, ldb.Message))
                count += 1
                msgs.append(me)
                break
            try:
                s.result()
                self.fail()
            except RuntimeError as re:
                pass
            for me in s:
                self.assertTrue(isinstance(me, ldb.Message))
                count += 1
                msgs.append(me)
                break
            for me in s:
                self.fail()

            r = s.result()
            self.assertEqual(len(r), 0)
            self.assertEqual(len(msgs), 2)
            if msgs[0].dn == m1.dn:
                self.assertEqual(msgs[0].dn, m1.dn)
                self.assertEqual(msgs[1].dn, m2.dn)
            else:
                self.assertEqual(msgs[0].dn, m2.dn)
                self.assertEqual(msgs[1].dn, m1.dn)
        finally:
            l.delete(ldb.Dn(l, "dc=foo4"))
            l.delete(ldb.Dn(l, "dc=foo5"))

    def test_add_text(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=foo4")
        m["bla"] = "bla"
        m["objectUUID"] = b"0123456789abcdef"
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        try:
            self.assertEqual(len(l.search()), 1)
        finally:
            l.delete(ldb.Dn(l, "dc=foo4"))

    def test_add_w_unhandled_ctrl(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=foo4")
        m["bla"] = b"bla"
        self.assertEqual(len(l.search()), 0)
        self.assertRaises(ldb.LdbError, lambda: l.add(m, ["search_options:1:2"]))

    def test_add_dict(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = {"dn": ldb.Dn(l, "dc=foo5"),
             "bla": b"bla",
             "objectUUID": b"0123456789abcdef"}
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        try:
            self.assertEqual(len(l.search()), 1)
        finally:
            l.delete(ldb.Dn(l, "dc=foo5"))

    def test_add_dict_text(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = {"dn": ldb.Dn(l, "dc=foo5"),
             "bla": "bla",
             "objectUUID": b"0123456789abcdef"}
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        try:
            self.assertEqual(len(l.search()), 1)
        finally:
            l.delete(ldb.Dn(l, "dc=foo5"))

    def test_add_dict_string_dn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = {"dn": "dc=foo6", "bla": b"bla",
             "objectUUID": b"0123456789abcdef"}
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        try:
            self.assertEqual(len(l.search()), 1)
        finally:
            l.delete(ldb.Dn(l, "dc=foo6"))

    def test_add_dict_bytes_dn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = {"dn": b"dc=foo6", "bla": b"bla",
             "objectUUID": b"0123456789abcdef"}
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        try:
            self.assertEqual(len(l.search()), 1)
        finally:
            l.delete(ldb.Dn(l, "dc=foo6"))

    def test_rename(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=foo7")
        m["bla"] = b"bla"
        m["objectUUID"] = b"0123456789abcdef"
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        try:
            l.rename(ldb.Dn(l, "dc=foo7"), ldb.Dn(l, "dc=bar"))
            self.assertEqual(len(l.search()), 1)
        finally:
            l.delete(ldb.Dn(l, "dc=bar"))

    def test_rename_string_dns(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=foo8")
        m["bla"] = b"bla"
        m["objectUUID"] = b"0123456789abcdef"
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        self.assertEqual(len(l.search()), 1)
        try:
            l.rename("dc=foo8", "dc=bar")
            self.assertEqual(len(l.search()), 1)
        finally:
            l.delete(ldb.Dn(l, "dc=bar"))

    def test_rename_bad_string_dns(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=foo8")
        m["bla"] = b"bla"
        m["objectUUID"] = b"0123456789abcdef"
        self.assertEqual(len(l.search()), 0)
        l.add(m)
        self.assertEqual(len(l.search()), 1)
        self.assertRaises(ldb.LdbError,lambda: l.rename("dcXfoo8", "dc=bar"))
        self.assertRaises(ldb.LdbError,lambda: l.rename("dc=foo8", "dcXbar"))
        l.delete(ldb.Dn(l, "dc=foo8"))

    def test_empty_dn(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertEqual(0, len(l.search()))
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=empty")
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        rm = l.search()
        self.assertEqual(1, len(rm))
        self.assertEqual(set(["dn", "distinguishedName", "objectUUID"]),
                         set(rm[0].keys()))

        rm = l.search(m.dn)
        self.assertEqual(1, len(rm))
        self.assertEqual(set(["dn", "distinguishedName", "objectUUID"]),
                         set(rm[0].keys()))
        rm = l.search(m.dn, attrs=["blah"])
        self.assertEqual(1, len(rm))
        self.assertEqual(0, len(rm[0]))

    def test_modify_delete(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=modifydelete")
        m["bla"] = [b"1234"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        rm = l.search(m.dn)[0]
        self.assertEqual([b"1234"], list(rm["bla"]))
        try:
            m = ldb.Message()
            m.dn = ldb.Dn(l, "dc=modifydelete")
            m["bla"] = ldb.MessageElement([], ldb.FLAG_MOD_DELETE, "bla")
            self.assertEqual(ldb.FLAG_MOD_DELETE, m["bla"].flags())
            l.modify(m)
            rm = l.search(m.dn)
            self.assertEqual(1, len(rm))
            self.assertEqual(set(["dn", "distinguishedName", "objectUUID"]),
                             set(rm[0].keys()))
            rm = l.search(m.dn, attrs=["bla"])
            self.assertEqual(1, len(rm))
            self.assertEqual(0, len(rm[0]))
        finally:
            l.delete(ldb.Dn(l, "dc=modifydelete"))

    def test_modify_delete_text(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=modifydelete")
        m.text["bla"] = ["1234"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        rm = l.search(m.dn)[0]
        self.assertEqual(["1234"], list(rm.text["bla"]))
        try:
            m = ldb.Message()
            m.dn = ldb.Dn(l, "dc=modifydelete")
            m["bla"] = ldb.MessageElement([], ldb.FLAG_MOD_DELETE, "bla")
            self.assertEqual(ldb.FLAG_MOD_DELETE, m["bla"].flags())
            l.modify(m)
            rm = l.search(m.dn)
            self.assertEqual(1, len(rm))
            self.assertEqual(set(["dn", "distinguishedName", "objectUUID"]),
                             set(rm[0].keys()))
            rm = l.search(m.dn, attrs=["bla"])
            self.assertEqual(1, len(rm))
            self.assertEqual(0, len(rm[0]))
        finally:
            l.delete(ldb.Dn(l, "dc=modifydelete"))

    def test_modify_add(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=add")
        m["bla"] = [b"1234"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        try:
            m = ldb.Message()
            m.dn = ldb.Dn(l, "dc=add")
            m["bla"] = ldb.MessageElement([b"456"], ldb.FLAG_MOD_ADD, "bla")
            self.assertEqual(ldb.FLAG_MOD_ADD, m["bla"].flags())
            l.modify(m)
            rm = l.search(m.dn)[0]
            self.assertEqual(3, len(rm))
            self.assertEqual([b"1234", b"456"], list(rm["bla"]))
        finally:
            l.delete(ldb.Dn(l, "dc=add"))

    def test_modify_add_text(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=add")
        m.text["bla"] = ["1234"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        try:
            m = ldb.Message()
            m.dn = ldb.Dn(l, "dc=add")
            m["bla"] = ldb.MessageElement(["456"], ldb.FLAG_MOD_ADD, "bla")
            self.assertEqual(ldb.FLAG_MOD_ADD, m["bla"].flags())
            l.modify(m)
            rm = l.search(m.dn)[0]
            self.assertEqual(3, len(rm))
            self.assertEqual(["1234", "456"], list(rm.text["bla"]))
        finally:
            l.delete(ldb.Dn(l, "dc=add"))

    def test_modify_replace(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=modify2")
        m["bla"] = [b"1234", b"456"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        try:
            m = ldb.Message()
            m.dn = ldb.Dn(l, "dc=modify2")
            m["bla"] = ldb.MessageElement([b"789"], ldb.FLAG_MOD_REPLACE, "bla")
            self.assertEqual(ldb.FLAG_MOD_REPLACE, m["bla"].flags())
            l.modify(m)
            rm = l.search(m.dn)[0]
            self.assertEqual(3, len(rm))
            self.assertEqual([b"789"], list(rm["bla"]))
            rm = l.search(m.dn, attrs=["bla"])[0]
            self.assertEqual(1, len(rm))
        finally:
            l.delete(ldb.Dn(l, "dc=modify2"))

    def test_modify_replace_text(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=modify2")
        m.text["bla"] = ["1234", "456"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        try:
            m = ldb.Message()
            m.dn = ldb.Dn(l, "dc=modify2")
            m["bla"] = ldb.MessageElement(["789"], ldb.FLAG_MOD_REPLACE, "bla")
            self.assertEqual(ldb.FLAG_MOD_REPLACE, m["bla"].flags())
            l.modify(m)
            rm = l.search(m.dn)[0]
            self.assertEqual(3, len(rm))
            self.assertEqual(["789"], list(rm.text["bla"]))
            rm = l.search(m.dn, attrs=["bla"])[0]
            self.assertEqual(1, len(rm))
        finally:
            l.delete(ldb.Dn(l, "dc=modify2"))

    def test_modify_flags_change(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=add")
        m["bla"] = [b"1234"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        try:
            m = ldb.Message()
            m.dn = ldb.Dn(l, "dc=add")
            m["bla"] = ldb.MessageElement([b"456"], ldb.FLAG_MOD_ADD, "bla")
            self.assertEqual(ldb.FLAG_MOD_ADD, m["bla"].flags())
            l.modify(m)
            rm = l.search(m.dn)[0]
            self.assertEqual(3, len(rm))
            self.assertEqual([b"1234", b"456"], list(rm["bla"]))

            # Now create another modify, but switch the flags before we do it
            m["bla"] = ldb.MessageElement([b"456"], ldb.FLAG_MOD_ADD, "bla")
            m["bla"].set_flags(ldb.FLAG_MOD_DELETE)
            l.modify(m)
            rm = l.search(m.dn, attrs=["bla"])[0]
            self.assertEqual(1, len(rm))
            self.assertEqual([b"1234"], list(rm["bla"]))
        finally:
            l.delete(ldb.Dn(l, "dc=add"))

    def test_modify_flags_change_text(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        m = ldb.Message()
        m.dn = ldb.Dn(l, "dc=add")
        m.text["bla"] = ["1234"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        try:
            m = ldb.Message()
            m.dn = ldb.Dn(l, "dc=add")
            m["bla"] = ldb.MessageElement(["456"], ldb.FLAG_MOD_ADD, "bla")
            self.assertEqual(ldb.FLAG_MOD_ADD, m["bla"].flags())
            l.modify(m)
            rm = l.search(m.dn)[0]
            self.assertEqual(3, len(rm))
            self.assertEqual(["1234", "456"], list(rm.text["bla"]))

            # Now create another modify, but switch the flags before we do it
            m["bla"] = ldb.MessageElement(["456"], ldb.FLAG_MOD_ADD, "bla")
            m["bla"].set_flags(ldb.FLAG_MOD_DELETE)
            l.modify(m)
            rm = l.search(m.dn, attrs=["bla"])[0]
            self.assertEqual(1, len(rm))
            self.assertEqual(["1234"], list(rm.text["bla"]))
        finally:
            l.delete(ldb.Dn(l, "dc=add"))

    def test_transaction_commit(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        l.transaction_start()
        m = ldb.Message(ldb.Dn(l, "dc=foo9"))
        m["foo"] = [b"bar"]
        m["objectUUID"] = b"0123456789abcdef"
        l.add(m)
        l.transaction_commit()
        l.delete(m.dn)

    def test_transaction_cancel(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        l.transaction_start()
        m = ldb.Message(ldb.Dn(l, "dc=foo10"))
        m["foo"] = [b"bar"]
        m["objectUUID"] = b"0123456789abcdee"
        l.add(m)
        l.transaction_cancel()
        self.assertEqual(0, len(l.search(ldb.Dn(l, "dc=foo10"))))

    def test_set_debug(self):
        def my_report_fn(level, text):
            pass
        l = ldb.Ldb(self.url(), flags=self.flags())
        l.set_debug(my_report_fn)

    def test_zero_byte_string(self):
        """Testing we do not get trapped in the \0 byte in a property string."""
        l = ldb.Ldb(self.url(), flags=self.flags())
        l.add({
            "dn": b"dc=somedn",
            "objectclass": b"user",
            "cN": b"LDAPtestUSER",
            "givenname": b"ldap",
            "displayname": b"foo\0bar",
            "objectUUID": b"0123456789abcdef"
        })
        res = l.search(expression="(dn=dc=somedn)")
        self.assertEqual(b"foo\0bar", res[0]["displayname"][0])

    def test_no_crash_broken_expr(self):
        l = ldb.Ldb(self.url(), flags=self.flags())
        self.assertRaises(ldb.LdbError, lambda: l.search("", ldb.SCOPE_SUBTREE, "&(dc=*)(dn=*)", ["dc"]))

# Run the SimpleLdb tests against an lmdb backend


class SimpleLdbLmdb(SimpleLdb):

    def setUp(self):
        if os.environ.get('HAVE_LMDB', '1') == '0':
            self.skipTest("No lmdb backend")
        self.prefix = MDB_PREFIX
        self.index = MDB_INDEX_OBJ
        super().setUp()


class SimpleLdbNoLmdb(LdbBaseTest):

    def setUp(self):
        if os.environ.get('HAVE_LMDB', '1') != '0':
            self.skipTest("lmdb backend enabled")
        self.prefix = MDB_PREFIX
        self.index = MDB_INDEX_OBJ
        super().setUp()

    def test_lmdb_disabled(self):
        self.testdir = tempdir()
        self.filename = os.path.join(self.testdir, "test.ldb")
        try:
            self.ldb = ldb.Ldb(self.url(), flags=self.flags())
            self.fail("Should have failed on missing LMDB")
        except ldb.LdbError as err:
            enum = err.args[0]
            self.assertEqual(enum, ldb.ERR_OTHER)


if __name__ == '__main__':
    import unittest
    unittest.TestProgram()



# waf build tool for building IDL files with pidl

import os, sys
from waflib import Build, Logs, Utils, Configure, Errors
from waflib.Configure import conf

@conf
def SAMBA_CHECK_PYTHON(conf, version=(3,6,0)):

    # enable tool to build python extensions
    if conf.env.HAVE_PYTHON_H:
        conf.check_python_version(version)
        return

    interpreters = []

    conf.find_program('python3', var='PYTHON',
                      mandatory=not conf.env.disable_python)
    conf.load('python')
    path_python = conf.find_program('python3')

    conf.env.PYTHON_SPECIFIED = (conf.env.PYTHON != path_python)
    conf.check_python_version(version)

    interpreters.append(conf.env['PYTHON'])
    conf.env.python_interpreters = interpreters


@conf
def SAMBA_CHECK_PYTHON_HEADERS(conf):
    if conf.env.disable_python:

        conf.msg("python headers", "Check disabled due to --disable-python")
        # we don't want PYTHONDIR in config.h, as otherwise changing
        # --prefix causes a complete rebuild
        conf.env.DEFINES = [x for x in conf.env.DEFINES
            if not x.startswith('PYTHONDIR=')
            and not x.startswith('PYTHONARCHDIR=')]

        return

    if conf.env["python_headers_checked"] == []:
        _check_python_headers(conf)
        conf.env["python_headers_checked"] = "yes"

    else:
        conf.msg("python headers", "using cache")

    # we don't want PYTHONDIR in config.h, as otherwise changing
    # --prefix causes a complete rebuild
    conf.env.DEFINES = [x for x in conf.env.DEFINES
        if not x.startswith('PYTHONDIR=')
        and not x.startswith('PYTHONARCHDIR=')]

def _check_python_headers(conf):
    conf.check_python_headers()

    abi_pattern = os.path.splitext(conf.env['pyext_PATTERN'])[0]
    conf.env['PYTHON_SO_ABI_FLAG'] = abi_pattern % ''
    conf.env['PYTHON_LIBNAME_SO_ABI_FLAG'] = (
        conf.env['PYTHON_SO_ABI_FLAG'].replace('_', '-'))

    for lib in conf.env['LINKFLAGS_PYEMBED']:
        if lib.startswith('-L'):
            conf.env.append_unique('LIBPATH_PYEMBED', lib[2:]) # strip '-L'
            conf.env['LINKFLAGS_PYEMBED'].remove(lib)

    # same as in waf 1.5, keep only '-fno-strict-aliasing'
    # and ignore defines such as NDEBUG _FORTIFY_SOURCE=2
    conf.env.DEFINES_PYEXT = []
    conf.env.CFLAGS_PYEXT = ['-fno-strict-aliasing']

    return

def PYTHON_BUILD_IS_ENABLED(self):
    return self.CONFIG_SET('HAVE_PYTHON_H')

Build.BuildContext.PYTHON_BUILD_IS_ENABLED = PYTHON_BUILD_IS_ENABLED


def SAMBA_PYTHON(bld, name,
                 source='',
                 deps='',
                 public_deps='',
                 realname=None,
                 cflags='',
                 cflags_end=None,
                 includes='',
                 init_function_sentinel=None,
                 local_include=True,
                 vars=None,
                 install=True,
                 enabled=True):
    '''build a python extension for Samba'''

    # force-disable when we can't build python modules, so
    # every single call doesn't need to pass this in.
    if not bld.PYTHON_BUILD_IS_ENABLED():
        enabled = False

    # Save time, no need to build python bindings when fuzzing
    if bld.env.enable_fuzzing:
        enabled = False

    # when we support static python modules we'll need to gather
    # the list from all the SAMBA_PYTHON() targets
    if init_function_sentinel is not None:
        cflags += ' -DSTATIC_LIBPYTHON_MODULES=%s' % init_function_sentinel

    # From https://docs.python.org/2/c-api/arg.html:
    # Starting with Python 2.5 the type of the length argument to
    # PyArg_ParseTuple(), PyArg_ParseTupleAndKeywords() and PyArg_Parse()
    # can be controlled by defining the macro PY_SSIZE_T_CLEAN before
    # including Python.h. If the macro is defined, length is a Py_ssize_t
    # rather than an int.

    # Because <Python.h> is often included before includes.h/config.h
    # This must be in the -D compiler options
    cflags += ' -DPY_SSIZE_T_CLEAN=1'

    source = bld.EXPAND_VARIABLES(source, vars=vars)

    if realname is not None:
        link_name = 'python/%s' % realname
    else:
        link_name = None

    bld.SAMBA_LIBRARY(name,
                      source=source,
                      deps=deps,
                      public_deps=public_deps,
                      includes=includes,
                      cflags=cflags,
                      cflags_end=cflags_end,
                      local_include=local_include,
                      vars=vars,
                      realname=realname,
                      link_name=link_name,
                      pyext=True,
                      target_type='PYTHON',
                      install_path='${PYTHONARCHDIR}',
                      allow_undefined_symbols=True,
                      install=install,
                      enabled=enabled)

Build.BuildContext.SAMBA_PYTHON = SAMBA_PYTHON


def pyembed_libname(bld, name):
    if bld.env['PYTHON_SO_ABI_FLAG']:
        return name + bld.env['PYTHON_SO_ABI_FLAG']
    else:
        return name

Build.BuildContext.pyembed_libname = pyembed_libname
